// efg, October 1998

// All of the work below is to determine when the cursor exits the TImage
// so the pixel data can be blanked (seems like more work than it's worth).
// The alternative is to use OnMouseMove events on the Form and other
// components to catch the cursor moving outside the TImage area.  This
// can work sometimes but is not fool proof.

// Here, explore use of modified TImage without modifying the component pallete.
// "Pollution" in the Component Palette seems to be common and should be
// minimized.  A good development environment should not use specialty
// components added for small, trivial modifications.  Maintenance costs
// over the software life cycle (and over a Delphi upgrade) can be severe.
// Keeping a common working component palette for a development group
// seems to be impossible.  (Which would mean you could never get past
// the intial "chaotic" Software Maturity level.)

// But the "non-visual" kludge used here isn't that good either.  The real
// solution for what I want here (a MouseEnter and MouseExit event for a
// TImage component would be for Inprise to make a better TImage VCL
// component.  With almost no chance of that happening, both alternatives
// seem to be bad.

// Ideas below were adapted from www.openinfo.se/martin/texterdelphi/mouseex.htm

UNIT ModifiedTImage;

INTERFACE

  USES
    Messages,  // TMessage
    Classes,   // TComponent
    Controls,  // cm_MouseEnter
    ExtCtrls;  // TImage

  TYPE
    TMouseInImage = (miiUnknown, miiYes, miiNo);
    TMouseEnterProcedure = PROCEDURE;
    TMouseLeaveProcedure = PROCEDURE;

    // A "Sensitive" image "knows" when the mouse enters or exits.
    TSensitiveImage =
    CLASS(TImage)
      PRIVATE
        FMouseInImage       :  TMouseInImage;
        FMouseEnterProcedure:  TMouseEnterProcedure;
        FMouseLeaveProcedure:  TMouseLeaveProcedure;

      PUBLIC
        CONSTRUCTOR Create (AOwner:  TComponent);  OverRide;

        PROCEDURE CMMouseEnter(VAR Message:  TMessage); message cm_MouseEnter;
        PROCEDURE CMMouseLeave(VAR Message:  TMessage); message cm_MouseLeave;

        PROPERTY MouseInImage:  TMouseInImage        READ FMouseInImage;
        PROPERTY OnMouseEnter:  TMouseEnterProcedure WRITE FMouseEnterProcedure;
        PROPERTY OnMouseLeave:  TMouseLeaveProcedure WRITE FMouseLeaveProcedure;
    END;


IMPLEMENTATION

  CONSTRUCTOR TSensitiveImage.Create(AOwner:  TComponent);
  BEGIN
    Inherited Create(AOwner);
    FMouseInImage := miiUnknown;
    FMouseEnterProcedure := NIL;
    FMouseLeaveProcedure := NIL;
  END {Create};

  PROCEDURE TSensitiveImage.CMMouseEnter(VAR Message:  TMessage);
  BEGIN
    // patterned after TSpeedButton.CMMouseEnter
    IF   Enabled AND (DragMode <> dmAutomatic)
    THEN BEGIN
      FMouseInImage := miiYes;
      IF  Assigned(FMouseEnterProcedure)
      THEN FMouseEnterProcedure;
      Message.Result := 1
    END
  END;

  PROCEDURE TSensitiveImage.CMMouseLeave(VAR Message:  TMessage);
  BEGIN
    FMouseInImage := miiNo;
    IF  Assigned(FMouseLeaveProcedure)
    THEN FMouseLeaveProcedure;
    Message.Result := 1
  END;


END.
