// Allow selection of part of image that matches desired criteria based on
// values of R, G, B, H, S and/or V

// efg, October 1998

unit ScreenColorRange;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtDlgs, StdCtrls, Spin, ExtCtrls,
  ModifiedTImage;       // TSensitiveImage;

type
  TFormColorRange = class(TForm)
    ButtonLoad: TButton;
    OpenPictureDialog: TOpenPictureDialog;
    CheckBoxStretch: TCheckBox;
    GroupBoxSelection: TGroupBox;
    CheckBoxRed: TCheckBox;
    SpinEditMinRed: TSpinEdit;
    SpinEditMaxRed: TSpinEdit;
    CheckBoxGreen: TCheckBox;
    SpinEditMinGreen: TSpinEdit;
    SpinEditMaxGreen: TSpinEdit;
    CheckBoxSaturation: TCheckBox;
    SpinEditMinSaturation: TSpinEdit;
    SpinEditMaxSaturation: TSpinEdit;
    CheckBoxHue: TCheckBox;
    SpinEditMinHue: TSpinEdit;
    SpinEditMaxHue: TSpinEdit;
    CheckBoxBlue: TCheckBox;
    SpinEditMinBlue: TSpinEdit;
    SpinEditMaxBlue: TSpinEdit;
    LabelSelectTo: TLabel;
    RadioButtonInclude: TRadioButton;
    RadioButtonExclude: TRadioButton;
    LabelExcludeColor: TLabel;
    ShapeExcludeColor: TShape;
    ColorDialog: TColorDialog;
    LabelGreaterEqual: TLabel;
    LabelLessEqual: TLabel;
    CheckBoxValue: TCheckBox;
    SpinEditMinValue: TSpinEdit;
    SpinEditMaxValue: TSpinEdit;
    Bevel1: TBevel;
    LabelInternalComment: TLabel;
    LabelCoordinates: TLabel;
    LabelCount: TLabel;
    ButtonSave: TButton;
    ButtonUpdate: TButton;
    SavePictureDialog: TSavePictureDialog;
    LabelColorCount: TLabel;
    procedure ButtonLoadClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure CheckBoxStretchClick(Sender: TObject);
    procedure ShapeExcludeColorMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure ParameterChange(Sender: TObject);
    procedure ButtonUpdateClick(Sender: TObject);
    procedure ButtonSaveClick(Sender: TObject);
  private
    BitmapOriginal:  TBitmap;
    SensitiveImage:  TSensitiveImage;
    TotalColors   :  INTEGER;
    PROCEDURE UpdateDisplay;
    PROCEDURE UpdateStatistics;
    PROCEDURE SensitiveImageMouseMove(Sender: TObject;
                 Shift: TShiftState; X, Y: Integer);
  public
    { Public declarations }
  end;

var
  FormColorRange: TFormColorRange;

implementation
{$R *.DFM}
  USES
    IniFiles,                    // TIniFile
    ImageProcessingPrimitives,   // pRGBTripleArray;
    ColorLibrary,                // RGBTripleToHSV
    Math,                        // MinIntValue
    GraphicsConversionsLibrary;  // LoadGraphicsFile

  CONST
    KeywordSetup     = 'Setup';
    KeywordDirectory = 'Directory';

// Why can't this be part of TForm?
PROCEDURE BlankCaption;
BEGIN
  FormColorRange.LabelCoordinates.Caption := ''
END {BlankCaption};


procedure TFormColorRange.ButtonLoadClick(Sender: TObject);
  VAR
    Filename:  STRING;
    IniFile :  TIniFile;
    NewPath :  STRING;
begin
  IF   OpenPictureDialog.Execute
  THEN BEGIN
    // Load BitmapOriginal
    IF   Assigned(BitmapOriginal)
    THEN BitmapOriginal.Free;
    BitmapOriginal := LoadGraphicsFile(OpenPictureDialog.Filename);
    IF   BitmapOriginal.PixelFormat <> pf24bit
    THEN BitmapOriginal.PixelFormat := pf24bit;  // force this

    TotalColors := CountColors(BitmapOriginal);

    FormColorRange.Caption := 'Color Range:  ' + OpenPictureDialog.Filename;

    // Update INI file for next time
    Filename := ChangeFileExt(ParamStr(0), '.INI');
    NewPath := ExtractFilePath(OpenPictureDialog.FileName);
    OpenPictureDialog.InitialDir := NewPath;
    IniFile := TIniFile.Create(Filename);
    TRY
      IniFile.WriteString(KeywordSetup, KeywordDirectory, NewPath)
    FINALLY
      IniFile.Free
    END;
    // Flush INI cache
    WritePrivateProfileString(NIL, NIL, NIL, pChar(Filename));

    UpdateStatistics;
    UpdateDisplay;

    GroupBoxSelection.Visible := TRUE;
    ButtonSave.Visible := TRUE
  END
end;

procedure TFormColorRange.FormCreate(Sender: TObject);
  VAR
    IniFile:  TIniFile;
begin
  BitmapOriginal := TBitmap.Create;
  BitmapOriginal.Width  := Bevel1.Width  - 2;
  BitmapOriginal.Height := Bevel1.Height - 2;
  BitmapOriginal.PixelFormat := pf24bit;

  // Use bevel to define location of TSensitiveImage
  SensitiveImage := TSensitiveImage.Create(FormColorRange);
  WITH SensitiveImage DO
  BEGIN
    Parent := FormColorRange;
    // Bevel is one pixel around the TSensitiveImage
    Top    := Bevel1.Top  + 1;
    Left   := Bevel1.Left + 1;
    Width  := BitmapOriginal.Width;
    Height := BitmapOriginal.Height;
    OnMouseMove  := SensitiveImageMouseMove;
    OnMouseLeave := BlankCaption;
    Show
  END;

  //  System Font -- not sure why this doesn't work in design mode
  LabelGreaterEqual.Caption := #242;
  LabelLessEqual.Caption    := #243;

  IniFile := TIniFile.Create(ChangeFileExt(ParamStr(0), '.INI'));
  TRY
    OpenPictureDialog.InitialDir := IniFile.ReadString(KeywordSetup,
                                                       KeywordDirectory,
                                        ExtractFilePath(ParamStr(0)));
  FINALLY
    IniFile.Free
  END;

end;


procedure TFormColorRange.FormDestroy(Sender: TObject);
begin
  BitmapOriginal.Free
end;


procedure TFormColorRange.CheckBoxStretchClick(Sender: TObject);
begin
  SensitiveImage.Stretch := CheckBoxStretch.Checked
end;

procedure TFormColorRange.ShapeExcludeColorMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  IF   ColorDialog.Execute
  THEN BEGIN
    ShapeExcludeColor.Brush.Color := ColorDialog.Color;
    ButtonUpdate.Enabled := TRUE;
  END
end;

procedure TFormColorRange.ParameterChange(Sender: TObject);
begin
  ButtonUpdate.Enabled := TRUE;

  IF   Sender IS  TCheckBox
  THEN BEGIN
    IF   (Sender AS TCheckBox).Tag = CheckBoxRed.Tag
    THEN BEGIN
      SpinEditMinRed.Visible := CheckBoxRed.Checked;
      SpinEditMaxRed.Visible := CheckBoxRed.Checked
    END;

    IF   (Sender AS TCheckBox).Tag = CheckBoxGreen.Tag
    THEN BEGIN
      SpinEditMinGreen.Visible := CheckBoxGreen.Checked;
      SpinEditMaxGreen.Visible := CheckBoxGreen.Checked
    END;

    IF   (Sender AS TCheckBox).Tag = CheckBoxBlue.Tag
    THEN BEGIN
      SpinEditMinBlue.Visible := CheckBoxBlue.Checked;
      SpinEditMaxBlue.Visible := CheckBoxBlue.Checked
    END;

    IF   (Sender AS TCheckBox).Tag = CheckBoxHue.Tag
    THEN BEGIN
      SpinEditMinHue.Visible := CheckBoxHue.Checked;
      SpinEditMaxHue.Visible := CheckBoxHue.Checked
    END;

    IF   (Sender AS TCheckBox).Tag = CheckBoxSaturation.Tag
    THEN BEGIN
      SpinEditMinSaturation.Visible := CheckBoxSaturation.Checked;
      SpinEditMaxSaturation.Visible := CheckBoxSaturation.Checked
    END;

    IF   (Sender AS TCheckBox).Tag = CheckBoxValue.Tag
    THEN BEGIN
      SpinEditMinValue.Visible := CheckBoxValue.Checked;
      SpinEditMaxValue.Visible := CheckBoxValue.Checked
    END;

  END
end;



PROCEDURE TFormColorRange.UpdateDisplay;
  VAR
    Bitmap         :  TBitmap;

    CheckBlue      :  BOOLEAN;
    CheckGreen     :  BOOLEAN;
    CheckHue       :  BOOLEAN;
    CheckInclude   :  BOOLEAN;
    CheckRed       :  BOOLEAN;
    CheckSaturation:  BOOLEAN;
    CheckValue     :  BOOLEAN;

    CountExclude   :  INTEGER;
    CountTotal     :  INTEGER;

    ExcludeTriple  :  TRGBTriple;
    H              :  INTEGER;
    i              :  INTEGER;
    j              :  INTEGER;
    maxRed         :  INTEGER;
    maxGreen       :  INTEGER;
    maxBlue        :  INTEGER;
    maxHue         :  INTEGER;
    maxHue2        :  INTEGER;
    maxSaturation  :  INTEGER;
    maxValue       :  INTEGER;

    minRed         :  INTEGER;
    minGreen       :  INTEGER;
    minBlue        :  INTEGER;
    minHue         :  INTEGER;
    minHue2        :  INTEGER;
    minSaturation  :  INTEGER;
    minValue       :  INTEGER;
    OK             :  BOOLEAN;
    rowIn          :  pRGBTripleArray;
    rowOut         :  pRGBTripleArray;
    S              :  INTEGER;
    V              :  INTEGER;
BEGIN
  Screen.Cursor := crHourGlass;
  TRY
    Bitmap := TBitmap.Create;
    TRY
      Bitmap.Width  := BitmapOriginal.Width;
      Bitmap.Height := BitmapOriginal.Height;
      Bitmap.PixelFormat := BitmapOriginal.PixelFormat;

      // Use in-memory variables to avoid overhead of accessing VCL components
      // for simple value, especially the spinboxes!
      CheckRed        := CheckBoxRed.Checked;
      CheckGreen      := CheckBoxGreen.Checked;
      CheckBlue       := CheckBoxBlue.Checked;
      CheckHue        := CheckBoxHue.Checked;
      CheckSaturation := CheckBoxSaturation.Checked;
      CheckValue      := CheckBoxValue.Checked;
      CheckInclude := RadioButtonInclude.Checked;

      minRed   := SpinEditMinRed.Value;
      maxRed   := SpinEditMaxRed.Value;

      minGreen := SpinEditMinGreen.Value;
      maxGreen := SpinEditMaxGreen.Value;

      minBlue  := SpinEditMinBlue.Value;
      maxBlue  := SpinEditMaxBlue.Value;

      minHue   := SpinEditMinHue.Value;
      maxHue   := SpinEditMaxHue.Value;

      IF   minHue < 0
      THEN BEGIN
        minHue2 := 360+minHue;
        maxHue2 := 360;
        minHue  := 0
      END
      ELSE BEGIN
        minHue2 := 0;
        maxHue2 := 0
      END;

      minSaturation := SpinEditMinSaturation.Value;
      maxSaturation := SpinEditMaxSaturation.Value;

      minValue := SpinEditMinValue.Value;
      maxValue := SpinEditMaxValue.Value;

      ExcludeTriple := ColorToRGBTriple(ShapeExcludeColor.Brush.Color);

      CountTotal := Bitmap.Height * Bitmap.Width;
      CountExclude := 0;
      FOR j := 0 TO Bitmap.Height-1 DO
      BEGIN
        rowIn  := BitmapOriginal.Scanline[j];
        rowOut := Bitmap.Scanline[j];

        FOR i := 0 TO Bitmap.Width-1 DO
        BEGIN
          OK := TRUE;

          WITH rowIn[i] DO
          BEGIN
            IF   CheckRed
            THEN OK := OK AND (rgbtRed   >= minRed) AND
                              (rgbtRed   <= maxRed);

            IF   CheckGreen
            THEN OK := OK AND (rgbtGreen >= minGreen) AND
                              (rgbtGreen <= maxGreen);

            IF   CheckBlue
            THEN OK := OK AND (rgbtBlue  >= minBlue) AND
                              (rgbtBlue  <= maxBlue);

            IF   CheckHue OR CheckSaturation OR CHeckValue
            THEN BEGIN
              RGBTripleToHSV(RowIn[i], H,S,V);

              IF   CheckHue
              THEN
                IF   minHue2 = 0
                THEN OK := OK AND  (H  >= minHue)  AND (H <= maxHue)
                ELSE OK := OK AND (((H >= minHue)  AND (H <= maxHue)) OR
                                   ((H >= minHue2) AND (H <= maxHue2)));

              IF   CheckSaturation
              THEN OK := OK AND (S  >= minSaturation) AND
                                (S  <= maxSaturation);

              IF   CheckValue
              THEN OK := OK AND (V  >= minValue) AND
                                (V  <= maxValue);
            END

          END;

          IF   CheckInclude
          THEN BEGIN
            IF   OK
            THEN RowOut[i] := RowIn[i]
            ELSE BEGIN
              RowOut[i] := ExcludeTriple;
              INC (CountExclude)
            END;
          END
          ELSE BEGIN
            IF   OK
            THEN BEGIN
              RowOut[i] := ExcludeTriple;
              INC (CountExclude)
            END
            ELSE RowOut[i] := RowIn[i]
          END;

        END

      END;

      SensitiveImage.Picture.Graphic := Bitmap;
      LabelCount.Caption :=
           Format('%d included, %d excluded, %d total pixels; %d total colors',
                  [CountTotal-CountExclude,
                   CountExclude, CountTotal, TotalColors]);

      LabelColorCount.Caption := IntToStr( CountColors(Bitmap) ) + ' colors';

    FINALLY
      Bitmap.Free
    END

  FINALLY
    Screen.Cursor := crDefault
  END
END {UpdateDisplay};


PROCEDURE TFormColorRange.UpdateStatistics;
  VAR
    H              :  INTEGER;
    i              :  INTEGER;
    j              :  INTEGER;
    maxRed         :  INTEGER;
    maxGreen        :  INTEGER;
    maxBlue        :  INTEGER;
    maxHue         :  INTEGER;
    maxSaturation  :  INTEGER;
    maxValue       :  INTEGER;
    minRed         :  INTEGER;
    minGreen       :  INTEGER;
    minBlue        :  INTEGER;
    minHue         :  INTEGER;
    minSaturation  :  INTEGER;
    minValue       :  INTEGER;
    rowIn          :  pRGBTripleArray;
    S              :  INTEGER;
    V              :  INTEGER;
BEGIN
  minRed   := 255;
  maxRed   :=   0;

  minGreen := 255;
  maxGreen :=   0;

  minBlue  := 255;
  maxBlue  :=   0;

  minHue   := 360;
  maxHue   :=   0;

  minSaturation := 255;
  maxSaturation :=   0;

  minValue := 255;
  maxValue :=   0;

  FOR j := 0 TO BitmapOriginal.Height-1 DO
  BEGIN
    rowIn  := BitmapOriginal.Scanline[j];

    FOR i := 0 TO BitmapOriginal.Width-1 DO
    BEGIN
      WITH rowIn[i] DO
      BEGIN
        minRed   := MinIntValue([minRed,   rgbtRed]);
        minGreen := MinIntValue([minGreen, rgbtGreen]);
        minBlue  := MinIntValue([minBlue,  rgbtBlue]);

        maxRed   := MaxIntValue([maxRed,   rgbtRed]);
        maxGreen := MaxIntValue([maxGreen, rgbtGreen]);
        maxBlue  := MaxIntValue([maxBlue,  rgbtBlue]);

        RGBTripleToHSV(RowIn[i], H,S,V);
        minHue        := MinIntValue([minHue,        H]);
        minSaturation := MinIntValue([minSaturation, S]);
        minValue      := MinIntValue([minValue,      V]);

        maxHue        := MaxIntValue([maxHue,        H]);
        maxSaturation := MaxIntValue([maxSaturation, S]);
        maxValue      := MaxIntValue([maxValue,      V])
      END
    END

  END;

  SpinEditMinRed.Value   := minRed;
  SpinEditMinGreen.Value := minGreen;
  SpinEditMinBlue.Value  := minBlue;

  SpinEditMaxRed.Value   := maxRed;
  SpinEditMaxGreen.Value := maxGreen;
  SpinEditMaxBlue.Value  := maxBlue;

  SpinEditMinHue.Value        := minHue;
  SpinEditMinSaturation.Value := minSaturation;
  SpinEditMinValue.Value      := minValue;

  SpinEditMaxHue.Value        := maxHue;
  SpinEditMaxSaturation.Value := maxSaturation;
  SpinEditMaxValue.Value      := maxValue;
END {UpdateStatistics};


procedure TFormColorRange.ButtonUpdateClick(Sender: TObject);
begin
  UpdateDisplay;
  ButtonUpdate.Enabled := FALSE
end;


procedure TFormColorRange.SensitiveImageMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);

  VAR
    H,S,V    :  INTEGER;
    RGBTriple:  TRGBTriple;
    xActual  :  INTEGER;
    yActual  :  INTEGER;
begin
  IF  CheckBoxStretch.Checked
  THEN BEGIN
    xActual := MulDiv(X, BitmapOriginal.Width,  SensitiveImage.Width);
    yActual := MulDiv(Y, BitmapOriginal.Height, SensitiveImage.Height)
  END
  ELSE BEGIN
    xActual := X;
    yActual := Y
  END;

  IF   (xActual < BitmapOriginal.Width) AND
       (yActual < BitmapOriginal.Height)
  THEN BEGIN
    // Pixels is fast enough for MouseMove but is slow in general
    RGBTriple := pRGBTripleArray(BitmapOriginal.Scanline[yActual])[xActual];
    RGBTripleToHSV(RGBTriple, H,S,V);
    LabelCoordinates.Caption :=
      Format('(%d,%d)   RGB(%d,%d,%d)   HSV(%d,%d,%d)',
             [xActual, yActual,
              RGBTriple.rgbtRed, RGBTriple.rgbtGreen, RGBTriple.rgbtBlue,
              H,S,V])
  END
  ELSE LabelCoordinates.Caption := ''
end;

procedure TFormColorRange.ButtonSaveClick(Sender: TObject);
begin
  IF   SavePictureDialog.Execute
  THEN SensitiveImage.Picture.Bitmap.SaveToFile(SavePictureDialog.Filename)
end;

end.
